//!
//!

use std::str::FromStr;

use super::super::super::repositories;
use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
  pool: Pool<ConnectionManager<PgConnection>>,
  member_repository: repositories::member::repository::Repository,
  member_site_repository: repositories::member_site::repository::Repository,
  member_session_repository: repositories::member_session::repository::Repository,
  captcha_repository: repositories::captcha::repository::Repository,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Service of service.member.service.identity")
      .finish()
  }
}

impl Service {
  ///
  pub fn new(
    connection_broker: nats::asynk::Connection,
    queue_broker: String,
    pool: Pool<ConnectionManager<PgConnection>>,
  ) -> Service {
    Service {
      connection_broker,
      queue_broker,
      pool,
      member_repository: repositories::member::repository::Repository::new(),
      member_site_repository: repositories::member_site::repository::Repository::new(),
      member_session_repository: repositories::member_session::repository::Repository::new(),
      captcha_repository: repositories::captcha::repository::Repository::new(),
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(
      self.check_username_for_duplication(),
      self.check_nickname_for_duplication(),
      self.captcha(),
      self.signin(),
    )
    .map(|_| ())
  }

  async fn check_username_for_duplication(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::identity::SUBJECT_CHECK_USERNAME_FOR_DUPLICATION,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::identity::CheckUsernameForDuplicationRequest::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server {}", e),
          })
        })?;

        let m = self
          .member_repository
          .select_by_username(&conn, &req.username)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server {}", e),
            })
          })?;

        message
          .respond(
            bpr::ss::member::identity::CheckUsernameForDuplicationResponse {
              error: None,
              result: Some(
                bpr::ss::member::identity::check_username_for_duplication_response::Result {
                  duplicated: m.is_some(),
                },
              ),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::identity::CheckUsernameForDuplicationResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn check_nickname_for_duplication(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::identity::SUBJECT_CHECK_NICKNAME_FOR_DUPLICATION,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::identity::CheckNicknameForDuplicationRequest::decode(
          message.data.as_slice(),
        )
        .map_err(|e| {
          bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
            message: format!("invalid request: {}", e),
          })
        })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server {}", e),
          })
        })?;

        let m = self
          .member_repository
          .select_by_nickname(&conn, &req.nickname)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        message
          .respond(
            bpr::ss::member::identity::CheckNicknameForDuplicationResponse {
              error: None,
              result: Some(
                bpr::ss::member::identity::check_nickname_for_duplication_response::Result {
                  duplicated: m.is_some(),
                },
              ),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::identity::CheckNicknameForDuplicationResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn captcha(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::identity::SUBJECT_CAPTCHA,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::identity::CaptchaRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let site_url = match client.site_url {
          Some(site_url) => site_url,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid site_url information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client.site_url".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server {}", e),
          })
        })?;

        match self
          .member_site_repository
          .select_by_url(&conn, &site_url)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server {}", e),
            })
          })? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid site_url information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client.site_url".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::None,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let mut c = captcha::Captcha::new();

        let c = c
          .apply_filter(captcha::filters::Noise::new(0.1))
          .view(220, 120);

        let image_as_base64 = match c.as_base64() {
          Some(s) => s,
          None => {
            return Err(bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: "captcha image encoding error[base64]".to_string(),
            }));
          }
        };

        let security_code = c.chars_as_string();
        let expires_at = (chrono::Utc::now() + chrono::Duration::hours(2)).timestamp();

        let new_captcha = repositories::captcha::models::NewCaptcha {
          security_code,
          expires_at,
        };

        let inserted_captcha = self
          .captcha_repository
          .insert(&conn, &new_captcha)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        message
          .respond(
            bpr::ss::member::identity::CaptchaResponse {
              error: None,
              result: Some(bpr::ss::member::identity::captcha_response::Result {
                token: inserted_captcha.id.to_string(),
                image: image_as_base64,
              }),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::identity::CaptchaResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn signin(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::identity::SUBJECT_SIGNIN,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::identity::SigninRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            message: format!("server {}", e),
          })
        })?;

        let ms = match self
          .member_site_repository
          .select_by_url(&conn, client.site_url())
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })? {
          Some(ms) => ms,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid site_url information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client.site_url".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::None,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let captcha_id = uuid::Uuid::from_str(req.token.as_str()).map_err(|e| {
          bcr::error::rpc::Error::InvalidParams(bcr::error::rpc::InvalidParams {
            message: "invalid captcha token".to_string(),
            detail: bcr::error::rpc::InvalidParamsDetail {
              location: "request".to_string(),
              param: "token".to_string(),
              value: "".to_string(),
              error_type: bcr::error::rpc::InvalidParamsType::None,
              message: e.to_string(),
            },
          })
        })?;
        let security_code = req.security_code;
        let username = req.username;
        let password = req.password;

        let captcha = match self
          .captcha_repository
          .select(&conn, captcha_id)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })? {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid captcha token".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "token".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::None,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        if !captcha.expires_at < chrono::Utc::now().timestamp() {
          return Err(bcr::error::rpc::Error::InvalidParams(
            bcr::error::rpc::InvalidParams {
              message: "invalid captcha token".to_string(),
              detail: bcr::error::rpc::InvalidParamsDetail {
                location: "request".to_string(),
                param: "token".to_string(),
                value: "".to_string(),
                error_type: bcr::error::rpc::InvalidParamsType::None,
                message: "captcha token is expired".to_string(),
              },
            },
          ));
        }

        if !captcha.security_code.eq(&security_code) {
          return Err(bcr::error::rpc::Error::InvalidParams(
            bcr::error::rpc::InvalidParams {
              message: "invalid security_code".to_string(),
              detail: bcr::error::rpc::InvalidParamsDetail {
                location: "request".to_string(),
                param: "token".to_string(),
                value: security_code.clone(),
                error_type: bcr::error::rpc::InvalidParamsType::EqualsTo,
                message: "security_code must equal to captcha".to_string(),
              },
            },
          ));
        }

        let m = match self
          .member_repository
          .select_by_username(&conn, &username)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })? {
          Some(m) => m,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid username".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "username".to_string(),
                  value: username,
                  error_type: bcr::error::rpc::InvalidParamsType::None,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        if m.member_site_id != ms.id {
          return Err(bcr::error::rpc::Error::InvalidParams(
            bcr::error::rpc::InvalidParams {
              message: "invalid site_url".to_string(),
              detail: bcr::error::rpc::InvalidParamsDetail {
                location: "header".to_string(),
                param: "client.site_url".to_string(),
                value: client.site_url().to_string(),
                error_type: bcr::error::rpc::InvalidParamsType::EqualsTo,
                message: "".to_string(),
              },
            },
          ));
        }

        if !(argon2::verify_encoded(&m.password, password.as_bytes()).map_err(|e| {
          bcr::error::rpc::Error::InvalidParams(bcr::error::rpc::InvalidParams {
            message: "invalid password".to_string(),
            detail: bcr::error::rpc::InvalidParamsDetail {
              location: "request".to_string(),
              param: "password".to_string(),
              value: password.clone(),
              error_type: bcr::error::rpc::InvalidParamsType::None,
              message: e.to_string(),
            },
          })
        })?) {
          return Err(bcr::error::rpc::Error::InvalidParams(
            bcr::error::rpc::InvalidParams {
              message: "invalid password".to_string(),
              detail: bcr::error::rpc::InvalidParamsDetail {
                location: "request".to_string(),
                param: "password".to_string(),
                value: password.clone(),
                error_type: bcr::error::rpc::InvalidParamsType::EqualsTo,
                message: "".to_string(),
              },
            },
          ));
        }

        let expires_at = (chrono::Utc::now() + chrono::Duration::minutes(30)).timestamp();
        let session = self
          .member_session_repository
          .insert(
            &conn,
            &repositories::member_session::models::NewMemberSession {
              member_id: m.id,
              ip: client.client_ip.clone(),
              expires_at,
            },
          )
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        message
          .respond(
            bpr::ss::member::identity::SigninResponse {
              error: None,
              result: Some(bpr::ss::member::identity::signin_response::Result {
                session_id: session.id.to_string(),
              }),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              message: format!("server: {}", e),
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::identity::SigninResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }
}
